{{
File: Servos.spin

Controls up to fourteen servos.

The top level program stores the pulse width information in a variable
(or an array of variables for multiple servos).  This object checks the
variable value (or each array element) and updates the pulse widths
transmitted to each servo every 20 ms.

v1.2 updated 6/13/07

Fixed glitch at 2.5 ms pulses.
Added support for up to 14 servos with 1 cog.
Added comments to servos method
}}

VAR

  ' Order matters for declarations of pinAddr, pulseAddr, servoCnt, pulseCnt
  ' because of longmove command in the start method.
  long pinAddr, pulseAddr, servoCnt, pulseCnt  
  long stack[30], cog


PUB start(pinAddress, pulseAddress, servoCount) : okay
{{
Launches servo control process into a cog.
   pinAddress   - address of servo I/O pin numbers array.
   pulseAddress - address of survo pulse durations array.
   servoCount   - number of servos (i.e. number of array elements).
   Returns t/f for success.
   IMPORTANT: The parent object should initialize the pin and pulse arrays
              before calling this start method.
}}
  servoCount--
  longmove(@pinAddr, @pinAddress, 3)
  okay := cog := cognew(servos, @stack) + 1
  if okay
    okay := @pulseCnt

PUB stop

  '' Stop servos object and free the cog

  if cog
    cogstop(cog~ - 1)

PUB servos | index, center, frame, mark, end, i, us

  frqa := frqb := 1                          ' freq increments phs by 1
  phsa := phsb := 0                          ' clear phs -> .bit31 pins low
  ctra[30..26] := ctrb[30..26] := %00100     ' Config counters to NCO

  repeat index from 0 to servoCnt            ' pins array elements -> output
    dira[byte[pinAddr][index]]~~

  us     := clkfreq / 1_000_000              ' 1 microsecond
  center := 1500 * us                        ' Center pulse = 1.5 ms
  frame  := 2700 * us                        ' Pulse frame to 20/7 ms
  end    := 1100 * us
  
  mark   := cnt                              ' Mark current cnt
  
  repeat
    i := -1
    pulseCnt++
    repeat
      if ++i =< servoCnt 
        ctra[8..0] := byte[pinAddr][i]
        phsa := -((long[pulseAddr][i] #> -1000 <# 1000 ) * us + center)     
      if ++i =< servoCnt                          
        ctrb[8..0] := byte[pinAddr][i]
        phsb := -((long[pulseAddr][i] #> -1000 <# 1000 ) * us + center)     
      waitcnt(mark += frame)                 
      if i == 13
        waitcnt(mark += end)
        quit
        